from tkinter import Canvas

class InvalidValueError(Exception):
    def __init__(self, value):
        self.value = value
        self.err = f"Invalid value {value}."
        super().__init__(self.err)


class InvalidPathError(Exception):
    def __init__(self, value):
        self.value = value
        self.err = f"Can't find path {value}."
        super().__init__(self.err)
        

class ResizingCanvas(Canvas):
    def __init__(self,parent,**kwargs):
        Canvas.__init__(self,parent,**kwargs)
        self.bind("<Configure>", self.on_resize)
        self.height = self.winfo_reqheight()
        self.width = self.winfo_reqwidth()

    def on_resize(self,event):
        # determine the ratio of old width/height to new width/height
        wscale = float(event.width)/self.width
        hscale = float(event.height)/self.height
        self.width = event.width
        self.height = event.height
        # resize the canvas 
        self.config(width=self.width, height=self.height)
        # rescale all the objects tagged with the "all" tag
        self.scale("all",0,0,wscale,hscale)


from time import sleep
from tkinter import BOTH, YES, TclError, Tk, PhotoImage

from realityengine.util import InvalidValueError, InvalidPathError, ResizingCanvas
from realityengine.game_object import gameobjects

import sys

class Window:
    def __init__(self, width: int, height: int, title: str = "RealityEngine Window", bg: str="#000000", FPS: int=15):
        """
        Creates the main window and sets up the canvas.\n
        \n
        `title` - The main title of the window.\n
        `width` - The Width of the window in pixels.\n
        `height` -  The height of the window in pixels.
        """
        self.title = title
        self.width = width
        self.height = height
        self.bg = bg
        self.FPS = FPS

        self.window = Tk()
        self.window.configure(background=bg)
        self.window.wm_title(title)
        self.window.geometry(f"{self.width}x{self.height}")
        self.window.protocol("WM_DELETE_WINDOW", self.close)

        self.canvas = ResizingCanvas(self.window, width=self.width, height=self.height, bg=self.bg, highlightthickness=0)
        self.canvas.pack(fill=BOTH, expand=YES)

        self.custom_update = None

        self.open = False

    def set_custom_update(self, method):
        """
        Sets the method to be run when update occurs to the `method` argument.

        Arguments:
            `method` | The function to be set as the custom update.
        """
        self.custom_update = method

    def resize(self, width: int, height: int):
        """
        Sets the size of the window.

        Arguments:
            `width`  | The new width of the window.\n
            `height` | The new height of the window.
        """
        self.window.geometry(f"{width}x{height}")
        self.width = width
        self.height = height

    def clock(self, fps):
        """
        Sets the clock speed of the window.

        Arguments:
            `fps` | The clock speed.
        """
        self.FPS = fps

    def run(self):
        """
        Starts the main window and sets up window events.
        """
        self.open = True
        self.window.after( self.FPS, self.update)
        self.window.mainloop()

    def clear(self):
        """
        Clears the window by un-rendering all drawn objects.
        """
        try:
            self.canvas.delete("all")
        except:
            pass

    def update(self):
        global run_self
        """
        Clears the screen and updates then draws all active gameobjects.
        """
        self.clear()

        if self.open == False:
            self.run()

        for go in gameobjects:
            go.update()
            transform = go.transform
            if go.shape == "rectangle":
                self.draw_rect(transform.top_corner[0], transform.top_corner[1], transform.bottom_corner[0], transform.bottom_corner[1], "", go.colour, go.layer)
            elif go.shape == "circle":
                self.draw_oval(transform.top_corner[0], transform.top_corner[1], transform.bottom_corner[0], transform.bottom_corner[1], 0, go.colour, go.layer)
            else:
                raise InvalidValueError(go.shape)

        if self.custom_update:
            self.custom_update()
        
        self.window.after(self.FPS, self.update)

    def set_title(self, title: str):
        """
        Sets the title of the window to the `title` argument.
        """
        self.window.wm_title(title)
        self.title = title

    def set_bg(self, bg: str):
        """
        Sets the background of the window to the `bg` argument.
        """
        try:
            self.window.configure(background=bg)
            self.canvas.configure(bg=bg)
            self.bg = bg
        except TclError:
            raise InvalidValueError(bg)

    def set_icon(self, icon_path: str):
        """
        Sets the icon of the window to the `icon_path` argument.
        """
        try:
            self.window.iconphoto(True, PhotoImage(file=icon_path))
        except TclError:
            raise InvalidPathError(icon_path)

    def set_resizable(self, resize_type: str):
        if resize_type == "none":
            self.window.resizable(False, False)
        elif resize_type == "x-axis":
            self.window.resizable(True, False)
        elif resize_type == "y-axis":
            self.window.resizable(False, True)
        elif resize_type == "both":
            self.window.resizable(True, True)

    def draw_rect(self, x_0: int, y_0: int, x_1: int, y_1: int, outline: str, colour: str, layer: str = "front"):
        try:
            rect = self.canvas.create_rectangle(x_0, y_0, x_1, y_1, fill=colour, outline=outline)
            if layer == "front":
                self.canvas.tag_raise(rect)
            elif layer == "back":
                self.canvas.tag_lower(rect)
        except TclError:
            pass

    def draw_oval(self, x_0: int, y_0: int, x_1: int, y_1: int, width: int, fill: str, layer: str = "front"):
        try:
            oval = self.canvas.create_oval(x_0, y_0, x_1, y_1, width=width, fill=fill)
            if layer == "front":
                self.canvas.tag_raise(oval)
            elif layer == "back":
                self.canvas.tag_lower(oval)
        except TclError:
            pass

    def draw_text(self, x: int, y: int, text: str, foreground: str = "black", font: tuple = ("Helvetica 15 bold"), layer: str = "front"):
        try:
            text = self.canvas.create_text(x, y, text=text, fill=foreground, font=font)
            if layer == "front":
                self.canvas.tag_raise(text)
            elif layer == "back":
                self.canvas.tag_lower(text)
        except TclError:
            pass
    
    def get_width(self):
        """
        Returns the width of the window in pixels.
        """
        return self.width

    def get_height(self):
        """
        Returns the height of the window in pixels.
        """
        return self.height
    
    def get_size(self):
        """
        Returns a tuple with the width and height of the window.
        """
        return (self.width, self.height)

    def close(self):
        self.window.destroy()
        sys.exit()

    def bind_key(self, key: str, method):
        """ 
        Binds the specified key to the given function and passes the event.

        Arguments:
            `key`    | The key that is binded to the method.\n
            `method` | The mthod that is run when the key is pressed.
        """
        self.window.bind(key, lambda event: method(event))

    def get_all_gameobjects_with_tag(self, tag: str):
        gos = []

        for go in gameobjects:
            if tag in go.tags:
                gos.append(go)

        return gos

    def halt_until_key(self, key: str):
        """
        Halt the program until the specified key is pressed.

        Arguments:
            `key` | The key that needs to be pressed.
        """
        halted = True

        def change(e):
            global halted
            if e.keysym == "<Return>":
                halted = False

        self.window.bind(key, lambda event: change)

        while not halted:
            sleep(0.2)


from setuptools import setup, find_packages

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Education",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS"
]

setup(
    name="realityengine",
    version="1.0.0",
    description="realityengine is a game library designed to make game development with Python easy.",
    long_description=open("README.txt").read() + "\n\n" + open("CHANGELOG.txt").read(),
    long_description_content_type="text/x-rst",
    url="",
    author="Nathaniel Chandler",
    author_email="nchand05@outlook.com",
    license="MIT",
    classifiers=classifiers,
    keywords="Python, Game Development, Educational",
    packages=find_packages(),
    install_requires=['simpleaudio', 'playsound']
)


# Simple Snake game in Python made with CatEngine
# This is a demo project designed to show some basic functionality of CatEngine


# ==// IMPORTS
from realityengine.window import Window
from realityengine.game_object import GameObject, Transform
from realityengine.audio import Sound

import random


#==// VARIABLES
SCREEN_WIDTH = 480
SCREEN_HEIGHT = 480

GRIDSIZE = 20
GRID_WIDTH = SCREEN_HEIGHT / GRIDSIZE
GRID_HEIGHT = SCREEN_WIDTH / GRIDSIZE

head_direction = "stop"
segments = []


#==// FUNCTIONS
def draw_grid():
    for y in range(0, int(GRID_HEIGHT)):
        for x in range(0, int(GRID_WIDTH)):
            if (x + y) % 2 == 0:
                rect = GameObject(Transform((x * GRIDSIZE) + (GRIDSIZE / 2), (y * GRIDSIZE) + (GRIDSIZE / 2), GRIDSIZE, GRIDSIZE, 1, 1), ["bg"])
                rect.colour = "#5DD8E4"
                rect.layer = "back"
                rect.start()
            else:
                rect = GameObject(Transform((x * GRIDSIZE) + (GRIDSIZE / 2), (y * GRIDSIZE) + (GRIDSIZE / 2), GRIDSIZE, GRIDSIZE, 1, 1), ["bg"])
                rect.colour = "#54C2CD"
                rect.layer = "back"
                rect.start()


def move():
    if head_direction == "up":
        snake_head.transform.change_position(snake_head.transform.x, snake_head.transform.y - 20)
    elif head_direction == "down":
        snake_head.transform.change_position(snake_head.transform.x, snake_head.transform.y + 20)
    elif head_direction == "left":
        snake_head.transform.change_position(snake_head.transform.x - 20, snake_head.transform.y)
    elif head_direction == "right":
        snake_head.transform.change_position(snake_head.transform.x + 20, snake_head.transform.y)


def go_up(e):
    global head_direction
    if head_direction != "down":
        head_direction = "up"


def go_down(e):
    global head_direction
    if head_direction != "up":
        head_direction = "down"


def go_left(e):
    global head_direction
    if head_direction != "right":
        head_direction = "left"


def go_right(e):
    global head_direction
    if head_direction != "left":
        head_direction = "right"


def end_game():
    global head_direction
    snake_head.transform.change_position(wn.get_width() / 2 + (GRIDSIZE / 2), wn.get_height() / 2 - (GRIDSIZE / 2))
    head_direction = "stop"

    for segment in segments:
        segment.destroy()

    segments.clear()

    food.transform.change_position(wn.get_width() / 2 + (GRIDSIZE / 2), wn.get_height() / 2 - (GRIDSIZE / 2) - 100)

    # Play the death sound
    dead = Sound("audio/death.wav")
    dead.play()


#==// GAME LOOP
def update():
    # Check for a collision with the border
    if snake_head.transform.is_touching_border(20, SCREEN_WIDTH, SCREEN_HEIGHT, 20):
        end_game()

    # Check for a collision with the food
    if snake_head.transform.distance(food.transform) < 20:
        # Move the food to a random spot
        x = (random.randint(0, GRID_WIDTH - 1) * GRIDSIZE) - (food.transform.width / 1.5)
        y = (random.randint(0, GRID_HEIGHT - 1) * GRIDSIZE) - (food.transform.height / 1.5)
        food.transform.change_position(x, y)

        # Add a segment
        new_segment_transform = Transform(wn.get_width() / 2 + (GRIDSIZE / 2), wn.get_height() / 2 - (GRIDSIZE / 2), 15, 15, 1, 1)
        new_segment = GameObject(new_segment_transform, ["snake", "snake_body"], "segment", "#11182F", "front")
        new_segment.start()
        segments.append(new_segment)

        # Play the collect sound
        collect = Sound("audio/food_collect.wav")
        collect.play()

    
    # Move the end segments first in reverse order
    for index in range(len(segments) - 1, 0, -1):
        x = segments[index - 1].transform.x
        y = segments[index - 1].transform.y
        segments[index].transform.change_position(x, y)

    # Move segment 0 to where the head is
    if len(segments) > 0:
        x = snake_head.transform.x
        y = snake_head.transform.y
        segments[0].transform.change_position(x, y)

    move()

    # Check for a head collision with the body segments
    for segment in segments:
        if segment.transform.distance(snake_head.transform) < 20:
            end_game()


#==// MAIN
if __name__ == "__main__":
    # Set up the screen
    wn = Window(SCREEN_WIDTH, SCREEN_HEIGHT, "Snake")
    wn.set_resizable("none")
    wn.clock(100)

    draw_grid()

    # Snake Head
    snake_head_transform = Transform(wn.get_width() / 2 + (GRIDSIZE / 2), wn.get_height() / 2 - (GRIDSIZE / 2), 15, 15, 1, 1)
    snake_head = GameObject(snake_head_transform, ["snake", "snake_head"], "snake_head", "#11182F", "front")
    snake_head.start()

    # Snake food
    food_transform = Transform(wn.get_width() / 2 + (GRIDSIZE / 2), wn.get_height() / 2 - (GRIDSIZE / 2) - 100, 15, 15, 1, 1)
    food = GameObject(food_transform, ["food"], "food", "#DFA331", "front")
    food.start()

    # Keyboard
    wn.bind_key("w", go_up)
    wn.bind_key("s", go_down)
    wn.bind_key("a", go_left)
    wn.bind_key("d", go_right)

    wn.set_custom_update(update)
    wn.run()

